#!/usr/bin/env python3
# simple check if two build or source graphs are equal
# instead of finding an isomorphism, we go through all nodes in g and check if
# the node is also in h and has the same predecessors and successors
# we then do the same for all nodes in h
# the algorithm is stupid but simple
# it is also faster than finding an isomorphism
# there might be a more complex, faster algorithm but most time is spent
# reading the graphs anyways

from __future__ import print_function
import sys
from collections import defaultdict
sys.path.append('/usr/share/botch')
from util import read_graphml

# given an attribute dict, returns a string describing this node
def stringofnode(attr):
    if attr['kind'] == "SCC":
        return "SCC (%s)"%(','.join(attr['sources']))
    elif attr['kind'] == "SrcPkg":
        return "SrcPkg (%s (= %s))"%(attr['name'], attr['version'])
    elif attr['kind'] == "InstSet":
        return "InstSet (%s (= %s): %s)"%(attr['name'], attr['version'], ','.join(attr['binaries']))
    else:
        raise Exception("unknown node kind: %s"%(attr['kind']), file=sys.stderr)

# given an attribute dict, returns a tuple uniquely identifying this node
def idofnode(attr):
    t = attr['kind']
    if t == "SrcPkg":
        return (t, attr['name'],attr['version'])
    elif t == "SCC":
        return (t, attr['sources'])
    elif t == "InstSet":
        return (t, attr['name'],attr['version'],attr['binaries'])
    else:
        raise Exception("unknown node kind: %s"%t, file=sys.stderr)

# given an id tuple, return a string describing it
def stringofid(attr):
    t = attr[0]
    if t == "SCC":
        return "SCC (%s)"%(','.join(attr[1]))
    elif t == "SrcPkg":
        return "SrcPkg (%s (= %s))"%(attr[1], attr[2])
    elif t == "InstSet":
        return "InstSet (%s (= %s): %s)"%(attr[1], attr[2], ','.join(attr[3]))
    else:
        raise Exception("unknown node kind: %s"%t, file=sys.stderr)

# given two node ids in g and h, check if they have the same predecessors and successors
def eqenv(g, h, gn, hn, verbose=False):
    gp = set([idofnode(g.node[n]) for n in g.predecessors(gn)])
    hp = set([idofnode(h.node[n]) for n in h.predecessors(hn)])
    gs = set([idofnode(g.node[n]) for n in g.successors(gn)])
    hs = set([idofnode(h.node[n]) for n in h.successors(hn)])
    if verbose and gp != hp:
        only_in_gp = gp - hp
        if only_in_gp:
            s = ",".join([stringofid(i) for i in only_in_gp])
            print("edge only in predecessors in g: %s"%s, file=sys.stderr)
        only_in_hp = hp - gp
        if only_in_hp:
            s = ",".join([stringofid(i) for i in only_in_hp])
            print("edge only in predecessors in h: %s"%s, file=sys.stderr)
    if verbose and gs != hs:
        only_in_gs = gs - hs
        if only_in_gs:
            s = ",".join([stringofid(i) for i in only_in_gs])
            print("edge only in successors in g: %s"%s, file=sys.stderr)
        only_in_hs = hs - gs
        if only_in_hs:
            s = ",".join([stringofid(i) for i in only_in_hs])
            print("edge only in successors in h: %s"%s, file=sys.stderr)
    return gp == hp \
       and gs == hs \

def graph_difference(g, h, verbose=False):
    # process both graphs and turn attribute strings representing sets into real sets
    # use frozenset because they will be used as dictionary keys and thus have to be hashable
    for n,attr in g.nodes_iter(data=True):
        if attr['kind'] == "SCC":
            g.node[n]['sources'] = frozenset([s for s in attr['sources'].split(',')])
        elif attr['kind'] == "InstSet":
            g.node[n]['binaries'] = frozenset([p for p in attr['binaries'].split(',')])

    for n,attr in h.nodes_iter(data=True):
        if attr['kind'] == "SCC":
            h.node[n]['sources'] = frozenset([s for s in attr['sources'].split(',')])
        elif attr['kind'] == "InstSet":
            h.node[n]['binaries'] = frozenset([p for p in attr['binaries'].split(',')])

    # build map from name/version tuple in h to node list
    h_map = defaultdict(list)
    for n,attr in h.nodes_iter(data=True):
        h_map[idofnode(attr)].append(n)

    # build map from name/version tuple in g to node list
    g_map = defaultdict(list)
    for n,attr in g.nodes_iter(data=True):
        g_map[idofnode(attr)].append(n)

    # we go through all nodes in g and check if they are all in h
    # then we go through all nodes in h and check if they are in g

    # for each list of nodes mapping to a package in g, check if there is a package
    # in h which connects to the same nodes
    for g_node,attr in g.nodes_iter(data=True):
        # find nodes in h that match the current node in g
        # if there is no such node, error out
        h_nodes = h_map[idofnode(attr)]
        if not h_nodes:
            if verbose:
                print("graphs differ", file=sys.stderr)
                print("cannot find %s in h"%stringofnode(attr), file=sys.stderr)
            return False
        # if there is one or more node, check if there is a node with the same
        # predecessors and successors
        for h_node in h_nodes:
            if eqenv(g, h, g_node, h_node, verbose):
                break
        else:
            if verbose:
                print("graphs differ", file=sys.stderr)
                print("cannot find %s in h with the same environment"%stringofnode(attr), file=sys.stderr)
            return False

    # for each list of nodes mapping to a package in h, check if there is a package
    # in g which connects to the same nodes
    for h_node,attr in h.nodes_iter(data=True):
        # find nodes in g that match the current node in h
        # if there is no such node, error out
        g_nodes = g_map[idofnode(attr)]
        if not g_nodes:
            if verbose:
                print("graphs differ", file=sys.stderr)
                print("cannot find %s in g"%stringofnode(attr), file=sys.stderr)
            return False
        # if there is one or more node, check if there is a node with the same
        # predecessors and successors
        for g_node in g_nodes:
            if eqenv(g, h, g_node, h_node, verbose):
                break
        else:
            if verbose:
                print("graphs differ", file=sys.stderr)
                print("cannot find %s in g with the same environment"%stringofnode(attr), file=sys.stderr)
            return False
    return True

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description=("Check if two graphs are the same"))
    parser.add_argument("g", type=read_graphml, help="graph g in GraphML format")
    parser.add_argument("h", type=read_graphml, help="graph h in GraphML format")
    parser.add_argument('-v', '--verbose', action='store_true', help='be verbose')
    args = parser.parse_args()
    ret = graph_difference(args.g, args.h, args.verbose)
    exit(not ret)
